Attribute VB_Name = "DBAccessionRecordHandler"
'
' ************************************************************************************************
' ************************************************************************************************
' ************************************************************************************************
' *****
' *****     Filename:   DBEntryDoc.bas
' *****     Name:       CuriousDB Accession Record Doc R/W (VB6)
' *****     Version:    1.0.0
' *****     Date:       30 Oct 2018
' *****     Author:     Alysson Rowan [alyssonrowan@gmail.com]
' *****
' *****     Copyright: Copyright (c) 2018, Alysson Rowan
' *****
' *****     This software is provided without warranty of any kind.
' *****     The user is responsible for testing and ensuring that the software is suitable for
' *****     the application it is included in.
' *****     No liability for any damages arising from the use of this software will be accepted
' *****     howsoever caused.
' *****
' *****     This program was created using only recycled electrons and magnetic domains.
' *****
' *****     Read the manual carefully before use.
' *****     The program is barely commented, so don't bother trying to read the comments.
' *****
' *****
' *****     Note:
' *****             FieldName:          Indicates the start of a new field
' *****             {  }                Indicate the start and end of a long text entry over multiple lines
' *****             Caption=            Indicates a named sub-field within the list of entries in a field
' *****                                 Absence of "Caption=" indicates a standard, unnamed sub-field
' *****             .name               Indicates a sub-property of a sub-field.
' *****                                 This name is terminated by whitespace.
' *****                                 The field is terminated by an end-of-line character.
' *****
' *****             Use spaces instead of tabs for whitespace.
' *****
' *****             NB: ALL field names are treated as arrays, resizing to suit.
' *****                 REDIM PRESERVE is your friend.
' *****

Public Type tAccessionSubField
    Index As String                         'Caption of sub-field
    Value As String                         'Value of field
    Type As Boolean                         'True indicates a multiline value.
    pSize As Long                           'Number of properties.
    pIndex() As String                      'Name of property.
    pValue() As String                      'Value of property.
End Type

Public Type tAccessionSubFieldArray         ' Array of values for a subfield.
    Size As Long                            'Size of array
    Index As String                         'Caption of sub-field - all of the values share the name!
    Value() As String                       'Value of field
    vType() As Boolean                      'True indicates a multiline value.
End Type

Public Type tAccessionField
    Caption As String                       'Field Name
    Entries As Long                         'Number of entries in field
    Entry() As tAccessionSubField           'Entries
End Type

Public Type tAccessionRecord
    RecordType As Long                      'Numerical code for mixed-recordtype databases
    RecordTitle As String                   'Record Type as a long name - use as document title?
    Timestamp As Date                       'Timestamp of last save update
    Fields As Long                          'Number of fields in record
    Field() As tAccessionField              'Data fields
End Type

Public RecordPath As String                 'Path to the record file
Public AutoBackup As Boolean                'Automatically backs up record file during write operation

Public Function FetchField(AccessionRecord As tAccessionRecord, FieldName As String) As tAccessionField
    
    Dim iLoop As Long
    ReDim FetchField.Entry(0)
    
    FetchField.Entries = 0
    FetchField.Caption = ""
    
    If AccessionRecord.Fields = 0 Then GoTo Done:
        
    For iLoop = 1 To AccessionRecord.Fields
    
        If AccessionRecord.Field(iLoop).Caption = FieldName Then
        
            ReDim FetchField.Entry(AccessionRecord.Field(iLoop).Entries)
            FetchField = AccessionRecord.Field(iLoop)
            Exit For
            
        End If
    
    Next iLoop

Done:

End Function

Public Function WriteField(AccessionRecord As tAccessionRecord, FieldName As String, FieldContent As tAccessionField) As tAccessionRecord
    Dim iLoop As Long
    
    WriteField = AccessionRecord
           
    For iLoop = 1 To AccessionRecord.Fields
    
        If AccessionRecord.Field(iLoop).Caption = FieldName Then
            
            WriteField.Field(iLoop) = FieldContent
            GoTo Done:
            
        End If
    
    Next iLoop
    
    ' dropped out the bottom, so it's a new field! => Append!
    
    ReDim Preserve WriteField.Field(WriteField.Fields + 1)
    WriteField.Fields = WriteField.Fields + 1
    WriteField.Field(WriteField.Fields) = FieldContent
    WriteField.Field(WriteField.Fields).Caption = FieldName ' update the field name.

Done:

End Function

Public Function FetchSubField(AccessionRecord As tAccessionRecord, FieldName As String, FieldSubName As String) As tAccessionSubFieldArray
    
    Dim WorkingField As tAccessionField
    Dim WorkingSubField As tAccessionSubFieldArray
    Dim iLoop As Long
    Dim vCount As Long
    
    ReDim WorkingSubField.Value(1)
    ReDim WorkingSubField.vType(1)
    ReDim FetchSubField.Value(1)
    ReDim FetchSubField.vType(1)
    
    FetchSubField.Size = 0
    FetchSubField.Index = ""
    FetchSubField.Value(1) = ""
    FetchSubField.vType(1) = False
    
    
    WorkingField = FetchField(AccessionRecord, FieldName)
    
    If WorkingField.Entries = 0 Then GoTo Done:
    
    For iLoop = 1 To WorkingField.Entries
        
        If WorkingField.Entry(iLoop).Index = FieldSubName Then
            
            vCount = 1
            WorkingSubField.Value(vCount) = WorkingField.Entry(iLoop).Value
            WorkingSubField.vType(vCount) = WorkingField.Entry(iLoop).Type
            WorkingSubField.Size = vCount
            WorkingSubField.Index = FieldSubName
            
            While (iLoop < WorkingField.Entries) And (WorkingField.Entry(iLoop + 1).Index = "")
                
                iLoop = iLoop + 1
                vCount = vCount + 1
                
                ReDim Preserve WorkingSubField.Value(vCount)
                ReDim Preserve WorkingSubField.vType(vCount)
                ReDim Preserve FetchSubField.Value(vCount)
                ReDim Preserve FetchSubField.vType(vCount)

                    
                WorkingSubField.Value(vCount) = WorkingField.Entry(iLoop).Value
                WorkingSubField.vType(vCount) = WorkingField.Entry(iLoop).Type
                WorkingSubField.Size = vCount
                    
            Wend
                
            
            iLoop = WorkingField.Entries + 1
            
        End If
        
    Next iLoop
    
    FetchSubField = WorkingSubField
        
Done:
    
End Function

Public Function WriteSubField(AccessionRecord As tAccessionRecord, FieldName As String, FieldSubName As String, SubFieldContent As tAccessionSubFieldArray) As Long

End Function

Public Function ReadRecord() As tAccessionRecord
    
    Dim FileNo As Long
    Dim FieldNo As Long
    Dim SubField As Long
    Dim LineBuffer As String
    Dim FieldTab As Long
    Dim FieldTabSpc As String
    Dim FieldValue As String
    Dim FieldWValue As String
    Dim FieldName As String
    Dim FieldWName As String
    Dim FieldSubName As String
    Dim FieldMultiline As Boolean
    Dim FieldMultiEntry As Boolean
    Dim Lines As Long
    Dim tmpRecord As tAccessionRecord
    Dim FieldSubProperty As Long
    
    FieldTab = 28
    FieldTabSpc = Space(28)
    FieldNo = 0
    SubField = 1
    ReDim tmpRecord.Field(1)
    
    
    
    FileNo = FreeFile
    
    If RecordPath = "" Then Exit Function
    
    Open RecordPath For Input As FileNo
    
    tmpRecord.RecordType = 0
    tmpRecord.RecordTitle = ""
    
    While Not EOF(FileNo)
        
        Line Input #FileNo, LineBuffer
        
        LineBuffer = Trim(LineBuffer)
        
        If LineBuffer = "" Then GoTo NextRecord:
        
        If Left(LineBuffer, 1) = "-" Then GoTo NextRecord:
        If Left(LineBuffer, 1) = "=" Then GoTo NextRecord:
        If Left(LineBuffer, 1) = "(" Then GoTo NextRecord:
        
        If Trim(Left(LineBuffer, 20)) = "RecordType is" Then
            tmpRecord.RecordType = Val(Trim(Mid(LineBuffer, 14, Len(LineBuffer))))
            GoTo NextRecord:
        End If
        
        
        If Trim(Left(LineBuffer, 20)) = "RecordTitle is" Then
            tmpRecord.RecordTitle = Trim(Mid(LineBuffer, 15, Len(LineBuffer)))
            GoTo NextRecord:
        End If
        
        If Trim(Left(LineBuffer, 20)) = "LastSavedTime is" Then
            tmpRecord.Timestamp = Trim(Mid(LineBuffer, 17, Len(LineBuffer)))
            GoTo NextRecord:
        End If
        
        If (InStr(1, LineBuffer, ":") = Len(LineBuffer)) Then   'next field
            
            FieldNo = FieldNo + 1
            SubField = 0
            FieldMultiline = False
            FieldMultiEntry = False
            FieldName = ""
            FieldSubName = ""
            FieldValue = ""
            
            ReDim Preserve tmpRecord.Field(FieldNo)
            ReDim Preserve tmpRecord.Field(FieldNo).Entry(1)
            ReDim Preserve tmpRecord.Field(FieldNo).Entry(1).pIndex(1)
            ReDim Preserve tmpRecord.Field(FieldNo).Entry(1).pValue(1)
            
            tmpRecord.Fields = FieldNo
            tmpRecord.Field(FieldNo).Caption = Trim(Left(LineBuffer, InStr(1, LineBuffer, ":") - 1))
            FieldName = Trim(Left(LineBuffer, InStr(1, LineBuffer, ":") - 1))
            Lines = Lines + 1
            
            GoTo NextRecord:
             
        End If
        
        
        If ((Left(LineBuffer, 1) = ".") And (FieldNo > 0) And (FieldMultiline = False)) = True Then   'found a subfield property
            ' SubField
            ' FieldSubProperty
            
            FieldSubProperty = tmpRecord.Field(FieldNo).Entry(SubField).pSize + 1
            ReDim Preserve tmpRecord.Field(FieldNo).Entry(SubField).pIndex(FieldSubProperty)
            ReDim Preserve tmpRecord.Field(FieldNo).Entry(SubField).pValue(FieldSubProperty)
            tmpRecord.Field(FieldNo).Entry(SubField).pSize = FieldSubProperty
            
            tmpRecord.Field(FieldNo).Entry(SubField).pIndex(FieldSubProperty) = Mid(LineBuffer, 2, InStr(2, LineBuffer, " ") - 1)
            tmpRecord.Field(FieldNo).Entry(SubField).pValue(FieldSubProperty) = Trim(Mid(LineBuffer, InStr(2, LineBuffer, " "), Len(LineBuffer)))
            
            GoTo NextRecord:
            
        End If
        
        If LineBuffer = "}" Then                                'end multi-line text
            'write the appropriate data to tmpRecord
            SubField = SubField + 1
            ReDim Preserve tmpRecord.Field(FieldNo).Entry(SubField)
            ReDim Preserve tmpRecord.Field(FieldNo).Entry(SubField).pIndex(1)
            ReDim Preserve tmpRecord.Field(FieldNo).Entry(SubField).pValue(1)
            tmpRecord.Field(FieldNo).Entry(SubField).pSize = 0


            tmpRecord.Field(FieldNo).Entry(SubField).Value = FieldValue
            tmpRecord.Field(FieldNo).Entry(SubField).Index = FieldSubName
            tmpRecord.Field(FieldNo).Entries = SubField
            tmpRecord.Field(FieldNo).Entry(SubField).Type = True
            
            FieldSubName = ""
            FieldValue = ""
            
            FieldMultiline = False
            Lines = Lines + 1

            GoTo NextRecord:
            
        End If
               
        If InStr(1, LineBuffer, "=") < FieldTab Then               'Subfield caption found
            ReDim Preserve tmpRecord.Field(FieldNo).Entry(SubField)
            
            If (InStr(1, LineBuffer, "{") < FieldTab) And (InStr(1, LineBuffer, "{") > 0) Then
                FieldMultiline = True
                FieldSubName = ""
                FieldValue = ""
                Lines = Lines + 1
            End If
                
            If InStr(1, LineBuffer, "=") > 0 Then
                FieldWName = Trim(Left(LineBuffer, InStr(1, LineBuffer, "=") - 1))
                FieldWValue = Trim(Mid(LineBuffer, InStr(1, LineBuffer, "=") + 1, Len(LineBuffer)))
            Else
                FieldWName = ""
                FieldWValue = Trim(LineBuffer)
            End If
        
            If FieldMultiline Then
                If Len(FieldValue) > 0 Then
                    FieldValue = FieldValue & vbCrLf
                End If
                FieldValue = Trim(FieldValue & Space(FieldTab) & Trim(Mid(LineBuffer, (InStr(1, LineBuffer, "{") + 1), Len(LineBuffer))))
                
                If FieldWName <> "" Then FieldSubName = FieldWName
            Else
                If FieldWValue <> "" Then
                    SubField = SubField + 1
                    ReDim Preserve tmpRecord.Field(FieldNo).Entry(SubField)
                    tmpRecord.Field(FieldNo).Entry(SubField).Index = FieldWName
                    tmpRecord.Field(FieldNo).Entry(SubField).Value = FieldWValue
                End If
            End If
            
            tmpRecord.Field(FieldNo).Entries = SubField

            Lines = Lines + 1
            GoTo NextRecord:
            
        End If
        
        
NextRecord:
    Wend
    tmpRecord.Field(fieldnumber).Entries = FieldNo
    
    Close FileNo
    
    ReadRecord = tmpRecord
    
End Function

Public Function WriteRecord(tmpRecord As tAccessionRecord) As Long
    
    
    Dim FileNo As Long
    Dim FileNo2 As Long
    Dim FieldNo As Long
    Dim SubField As Long
    Dim LineBuffer As String
    Dim FieldTab As Long
    Dim FieldTabSpc As String
    Dim FieldValue As String
    Dim FieldName As String
    Dim FieldMultiline As Boolean
    Dim FieldMultiEntry As Boolean
    Dim Lines As Long
    Dim pProperties As Long
    Dim tNow As Date
    
    FieldTab = 28
    FieldTabSpc = Space(28)
    FieldNo = 0
    SubField = 1
    
    
    FileNo = FreeFile
    
    
    If RecordPath = "" Then Exit Function
    
    If AutoBackup Then                          ' automatically backs up record file with an additional .bak extension
        On Error GoTo IsNewFile:
            Open RecordPath For Input As FileNo
            FileNo2 = FreeFile
            Open RecordPath & ".bak" For Output As FileNo2
            While Not EOF(FileNo)
                Line Input #FileNo, LineBuffer
                Print #FileNo2, LineBuffer
            Wend
IsNewFile:
            Close #FileNo
            Close #FileNo2
        On Error GoTo 0
    End If
            
        
    
    
    Open RecordPath For Output As FileNo
    
    If Len(tmpRecord.RecordTitle) > 0 Then
        Print #FileNo, Left("RecordTitle is" + FieldTabSpc, FieldTab); tmpRecord.RecordTitle
        Print #FileNo, FieldTabSpc; String(Len(tmpRecord.RecordTitle), "=")
        Print #FileNo,
    End If
    
    For fieldnumber = 1 To tmpRecord.Fields
        
        
        If fieldnumber > 1 Then ' separator before field group
            Print #FileNo, ""
            Print #FileNo, String(80, "=")
        End If
        Print #FileNo, Trim(tmpRecord.Field(fieldnumber).Caption); ":"
        Lines = Lines + 1
        
        For SubField = 1 To tmpRecord.Field(fieldnumber).Entries
           
            If Len(Trim(tmpRecord.Field(fieldnumber).Entry(SubField).Index)) > 0 Then
                If tmpRecord.Field(fieldnumber).Entry(SubField).Type Then
                    Print #FileNo, Left("  " + Trim(tmpRecord.Field(fieldnumber).Entry(SubField).Index) + " =" + FieldTabSpc, FieldTab - 2); "{ ";
                Else
                    Print #FileNo, Left("  " + Trim(tmpRecord.Field(fieldnumber).Entry(SubField).Index) + " =" + FieldTabSpc, FieldTab);
                End If
            Else
                Print #FileNo, Left(Trim(tmpRecord.Field(fieldnumber).Entry(SubField).Index) + FieldTabSpc, FieldTab);
            End If
            
            Print #FileNo, tmpRecord.Field(fieldnumber).Entry(SubField).Value
            
            If tmpRecord.Field(fieldnumber).Entry(SubField).Type Then
                Print #FileNo, Space(FieldTab - 2); "}"
            End If
            
            If tmpRecord.Field(fieldnumber).Entry(SubField).pSize > 0 Then
                For pProperties = 1 To tmpRecord.Field(fieldnumber).Entry(SubField).pSize
                    Print #FileNo, Left("    " + "." + Trim(tmpRecord.Field(fieldnumber).Entry(SubField).pIndex(pProperties)) + FieldTabSpc, FieldTab);
                    Print #FileNo, " "; tmpRecord.Field(fieldnumber).Entry(SubField).pValue(pProperties)
                Next pProperties
            End If
            
            Lines = Lines + 1
        Next SubField
    Next fieldnumber
        
NextRecord:

    tNow = Now()
    Print #FileNo, ""
    Print #FileNo, Left("===== DB System Info " + String(80, "="), 80)
    Print #FileNo, Left(" RecordType is" + FieldTabSpc, FieldTab); Trim(tmpRecord.RecordType)
    Print #FileNo, Left(" LastSavedTime is " + FieldTabSpc, FieldTab); tNow
    Print #FileNo, Left("(Previous Version was " + FieldTabSpc, FieldTab); tmpRecord.Timestamp; ")"
    Print #FileNo, String(80, "=")
    tmpRecord.Timestamp = tNow
    Close FileNo
    WriteRecord = 0
    
End Function

